"""Data conversion helper functions."""
from typing import Any, Callable, Optional


def nullable_convert(value: Any, func: Callable[[Any], Any]) -> Any:
    """Convert a value given a conversion function if it is not None.

    Remarks:
        - This will silently handle None inputs by returning None.
        - Useful for when a function depends on a non-None input.

    Args:
        value: The value to convert.
        func: The conversion function.

    Returns:
        The converted value, or None if the value is None.
    """
    if value is None:
        return value
    else:
        return func(value)


def int_or_none(value: Any) -> Optional[int]:
    """Convert a value to an integer, while silently handling a None value.

    Examples:
        >>> int_or_none('7')
        7
        >>> int_or_none(None) is None
        True
    """
    return nullable_convert(value, int)


def without_keys(d: dict, keys: set[Any]) -> dict:
    """Remove keys from a dict.

    >>> without_keys({'a': 1, 'b': 2}, {'a'})
    {'b': 2}
    >>> without_keys({'a': 1, 'b': 2}, {'c'}) == {'a': 1, 'b': 2}
    True
    """
    return {k: v for k, v in d.items() if k not in keys}



# -*- coding: utf-8 -*-
from ..marshmallow_mongoengine.conversion import fields


def _is_field(value):
    return isinstance(value, type) and issubclass(value, fields.Field)


class ModelConverter(object):
    """Class that converts a mongoengine Document into a dictionary of
    corresponding marshmallow `Fields <marshmallow.fields.Field>`.
    """

    def fields_for_model(self, model, fields_kwargs=None, fields=None):
        result = {}
        fields_kwargs = fields_kwargs or {}
        for field_name, field_me in model._fields.items():
            if fields and field_name not in fields:
                continue
            if field_name in fields_kwargs:
                field_ma_cls = self.convert_field(field_me, **fields_kwargs[field_name])
            else:
                field_ma_cls = self.convert_field(field_me)
            if field_ma_cls:
                result[field_name] = field_ma_cls
        return result

    def convert_field(self, field_me, instance=True, **kwargs):
        field_builder = fields.get_field_builder_for_data_type(field_me)
        if not instance:
            return field_builder.marshmallow_field_cls
        return field_builder.build_marshmallow_field(**kwargs)

    def field_for(self, model, property_name, **kwargs):
        field_me = getattr(model, property_name)
        field_builder = fields.get_field_builder_for_data_type(field_me)
        return field_builder.build_marshmallow_field(**kwargs)


default_converter = ModelConverter()


fields_for_model = default_converter.fields_for_model
"""Generate a dict of field_name: `marshmallow.fields.Field` pairs for the
given model.

:param model: The Mongoengine Document model
:return: dict of field_name: Field instance pairs
"""


convert_field = default_converter.convert_field
"""Convert a Mongoengine `Field` to a field instance or class.

:param Property field_me: Mongoengine Field Property.
:param fields_kwargs: Dict of per-field kwargs to pass at field creation.
:param bool instance: If `True`, return  `Field` instance, computing
    relevant kwargs from the given property. If `False`, return
    the `Field` class.
:param kwargs: Additional keyword arguments to pass to the field constructor.
:return: A `marshmallow.fields.Field` class or instance.
"""


field_for = default_converter.field_for
"""Convert a property for a mapped Mongoengine Document class to a
   marshmallow `Field`.
Example: ::

    date_created = field_for(Author, 'date_created', dump_only=True)
    author = field_for(Book, 'author')

:param type ,: A Mongoengine Document mapped class.
:param str property_name: The name of the property to convert.
:param kwargs: Extra keyword arguments to pass to `property2field`
:return: A `marshmallow.fields.Field` class or instance.
"""



import inspect
import functools

import mongoengine as me

from ...marshmallow_mongoengine import fields as ma_fields
from ...marshmallow_mongoengine.conversion import params
from ...marshmallow_mongoengine.exceptions import ModelConversionError


class MetaFieldBuilder(object):
    """
    Convert a given Mongoengine Field to an equivalent Marshmallow Field
    """

    BASE_AVAILABLE_PARAMS = (
        params.DescriptionParam,
        params.AllowNoneParam,
        params.ChoiceParam,
        params.RequiredParam,
    )
    AVAILABLE_PARAMS = ()
    MARSHMALLOW_FIELD_CLS = None

    def __init__(self, field):
        self.mongoengine_field = field
        self.params = [
            paramCls(field)
            for paramCls in self.BASE_AVAILABLE_PARAMS + self.AVAILABLE_PARAMS
        ]

    def build_marshmallow_field(self, **kwargs):
        """
        :return: The Marshmallow Field instanciated and configured
        """
        field_kwargs = None
        for param in self.params:
            field_kwargs = param.apply(field_kwargs)
        field_kwargs.update(kwargs)
        return self.marshmallow_field_cls(**field_kwargs)

    def _get_marshmallow_field_cls(self):
        """
        Return the marshmallow Field class, overload this method to
        generate more dynamic field class
        """
        return self.MARSHMALLOW_FIELD_CLS

    @property
    def marshmallow_field_cls(self):
        return self._get_marshmallow_field_cls()


class ListBuilder(MetaFieldBuilder):
    AVAILABLE_PARAMS = (params.LengthParam,)
    MARSHMALLOW_FIELD_CLS = ma_fields.List

    def _get_marshmallow_field_cls(self):
        sub_field = get_field_builder_for_data_type(self.mongoengine_field.field)
        return functools.partial(
            self.MARSHMALLOW_FIELD_CLS, sub_field.build_marshmallow_field()
        )


class EnumBuilder(MetaFieldBuilder):
    AVAILABLE_PARAMS = ()
    MARSHMALLOW_FIELD_CLS = ma_fields.Enum

    def _get_marshmallow_field_cls(self):
        return functools.partial(
            self.MARSHMALLOW_FIELD_CLS, self.mongoengine_field._enum_cls
        )


class ReferenceBuilder(MetaFieldBuilder):
    AVAILABLE_PARAMS = ()
    MARSHMALLOW_FIELD_CLS = ma_fields.Reference

    def _get_marshmallow_field_cls(self):
        return functools.partial(
            self.MARSHMALLOW_FIELD_CLS, self.mongoengine_field.document_type
        )


class GenericReferenceBuilder(MetaFieldBuilder):
    BASE_AVAILABLE_PARAMS = tuple(
        [
            p
            for p in MetaFieldBuilder.BASE_AVAILABLE_PARAMS
            if p is not params.ChoiceParam
        ]
    )
    AVAILABLE_PARAMS = ()
    MARSHMALLOW_FIELD_CLS = ma_fields.GenericReference

    def build_marshmallow_field(self, **kwargs):
        # Special handle for the choice field given it represent the
        # reference's document class
        kwargs["choices"] = getattr(self.mongoengine_field, "choices", None)
        return super(GenericReferenceBuilder, self).build_marshmallow_field(**kwargs)


class EmbeddedDocumentBuilder(MetaFieldBuilder):
    AVAILABLE_PARAMS = ()
    MARSHMALLOW_FIELD_CLS = ma_fields.Nested
    BASE_NESTED_SCHEMA_CLS = None

    def _get_marshmallow_field_cls(self):
        # Recursive build of marshmallow schema
        from ...marshmallow_mongoengine.schema import ModelSchema

        base_nested_schema_cls = self.BASE_NESTED_SCHEMA_CLS or ModelSchema

        class NestedSchema(base_nested_schema_cls):
            class Meta:
                model = self.mongoengine_field.document_type

        return functools.partial(self.MARSHMALLOW_FIELD_CLS, NestedSchema)


class MapBuilder(MetaFieldBuilder):
    AVAILABLE_PARAMS = ()
    MARSHMALLOW_FIELD_CLS = ma_fields.Map

    def _get_marshmallow_field_cls(self):
        # Recursive build of marshmallow schema
        from ...marshmallow_mongoengine.convert import convert_field

        return functools.partial(
            self.MARSHMALLOW_FIELD_CLS, convert_field(self.mongoengine_field.field)
        )


def get_field_builder_for_data_type(field_me):
    field_me_types = inspect.getmro(type(field_me))
    for field_me_type in field_me_types:
        if field_me_type in FIELD_MAPPING:
            field_ma_cls = FIELD_MAPPING[field_me_type]
            break
    else:
        raise ModelConversionError("Could not find field of type {0}.".format(field_me))
    return field_ma_cls(field_me)


FIELD_MAPPING = {}


def register_field_builder(mongo_field_cls, builder):
    """
    Register a :class MetaFieldBuilder: to a given Mongoengine Field
    :param mongo_field_cls: Mongoengine Field
    :param build: field_builder to register
    """
    FIELD_MAPPING[mongo_field_cls] = builder


def register_field(mongo_field_cls, marshmallow_field_cls, available_params=()):
    """
    Bind a marshmallow field to it corresponding mongoengine field
    :param mongo_field_cls: Mongoengine Field
    :param marshmallow_field_cls: Marshmallow Field
    :param available_params: List of :class marshmallow_mongoengine.conversion.params.MetaParam:
        instances to import the mongoengine field config to marshmallow
    """

    class Builder(MetaFieldBuilder):
        AVAILABLE_PARAMS = available_params
        MARSHMALLOW_FIELD_CLS = marshmallow_field_cls

    register_field_builder(mongo_field_cls, Builder)


register_field(me.fields.BinaryField, ma_fields.Raw)
register_field(me.fields.BooleanField, ma_fields.Boolean)
register_field(me.fields.ComplexDateTimeField, ma_fields.DateTime)
register_field(me.fields.DateField, ma_fields.Date)
register_field(me.fields.DateTimeField, ma_fields.DateTime)
register_field(
    me.fields.DecimalField,
    ma_fields.Decimal,
    available_params=(params.SizeParam, params.PrecisionParam),
)
register_field(me.fields.DictField, ma_fields.Raw)
register_field(me.fields.DynamicField, ma_fields.Raw)
register_field(
    me.fields.EmailField, ma_fields.Email, available_params=(params.LengthParam,)
)
register_field(
    me.fields.FloatField, ma_fields.Float, available_params=(params.SizeParam,)
)
register_field(
    me.fields.GenericEmbeddedDocumentField, ma_fields.GenericEmbeddedDocument
)
register_field_builder(me.fields.GenericReferenceField, GenericReferenceBuilder)
register_field_builder(me.fields.ReferenceField, ReferenceBuilder)
# LazyReferenceField and GenericLazyReference need mongoengine >= 0.15.0
if hasattr(me.fields, "LazyReferenceField"):
    register_field_builder(me.fields.LazyReferenceField, ReferenceBuilder)
if hasattr(me.fields, "GenericLazyReferenceField"):
    register_field_builder(me.fields.GenericLazyReferenceField, GenericReferenceBuilder)
# FilesField and ImageField can't be simply displayed...
register_field(me.fields.FileField, ma_fields.Skip)
register_field(me.fields.ImageField, ma_fields.Skip)
register_field(
    me.fields.IntField, ma_fields.Integer, available_params=(params.SizeParam,)
)
register_field(
    me.fields.LongField, ma_fields.Integer, available_params=(params.SizeParam,)
)
register_field(me.fields.ObjectIdField, ma_fields.ObjectId)
register_field(me.fields.UUIDField, ma_fields.UUID)
register_field(me.fields.PointField, ma_fields.Point)
register_field(me.fields.LineStringField, ma_fields.LineString)
register_field(
    me.fields.SequenceField, ma_fields.Integer, available_params=(params.SizeParam,)
)  # TODO: handle value_decorator
register_field(
    me.fields.StringField,
    ma_fields.String,
    available_params=(params.LengthParam, params.RegexParam),
)
register_field(
    me.fields.URLField, ma_fields.URL, available_params=(params.LengthParam,)
)
register_field_builder(me.fields.EmbeddedDocumentField, EmbeddedDocumentBuilder)
register_field_builder(me.fields.ListField, ListBuilder)
register_field_builder(me.fields.MapField, MapBuilder)
register_field_builder(me.fields.SortedListField, ListBuilder)
register_field_builder(me.fields.EnumField, EnumBuilder)
# TODO: finish fields...
# me.fields.GeoPointField: ma_fields.GeoPoint,
# me.fields.PolygonField: ma_fields.Polygon,
# me.fields.MultiPointField: ma_fields.MultiPoint,
# me.fields.MultiLineStringField: ma_fields.MultiLineString,
# me.fields.MultiPolygonField: ma_fields.MultiPolygon,



import bson
import uuid
from bson.errors import BSONError
from marshmallow import ValidationError, fields, missing
from mongoengine import ValidationError as MongoValidationError, NotRegistered
from mongoengine.base import get_document

# Republish default fields...
from marshmallow.fields import *  # noqa


class ObjectId(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return bson.ObjectId(value)
        except BSONError:
            raise ValidationError("invalid ObjectId `%s`" % value)

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return missing
        return str(value)


class Point(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return dict(
                type="Point", coordinates=[float(value["x"]), float(value["y"])]
            )
        except Exception:
            raise ValidationError("invalid Point `%s`" % value)

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return missing
        return dict(x=value["coordinates"][0], y=value["coordinates"][1])


class LineString(fields.Field):

    """
    Marshmallow custom field to map with :class Mongoengine.LineStringField:
    """

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            coordinates = []

            for x, y in value["coordinates"]:
                coordinates.append([float(x), float(y)])
            return dict(type="LineString", coordinates=coordinates)
        except Exception:
            raise ValidationError("invalid value data `%s`" % value)

    def _serialize(self, value, attr, obj):
        if value is None:
            return missing
        return dict(coordinates=value["coordinates"])


class Reference(fields.Field):

    """
    Marshmallow custom field to map with :class Mongoengine.ReferenceField:
    """

    def __init__(self, document_type_obj, *args, **kwargs):
        self.document_type_obj = document_type_obj
        super(Reference, self).__init__(*args, **kwargs)

    @property
    def document_type(self):
        if isinstance(self.document_type_obj, str):
            self.document_type_obj = get_document(self.document_type_obj)
        return self.document_type_obj

    def _deserialize(self, value, attr, data, **kwargs):
        document_type = self.document_type
        try:
            if isinstance(value, dict):
                return document_type(**value)

            return document_type.objects.get(pk=value)
        except (
            document_type.DoesNotExist,
            MongoValidationError,
            ValueError,
            TypeError,
        ):
            raise ValidationError(
                "unknown document %s `%s`" % (document_type._class_name, value)
            )
        return value

    def _serialize(self, value, attr, obj, **kwargs):
        # Only return the pk of the document for serialization
        if value is None:
            return missing
        return (
            str(value.pk)
            if isinstance(value.pk, (bson.ObjectId, uuid.UUID))
            else value.pk
        )


class GenericReference(fields.Field):

    """
    Marshmallow custom field to map with :class Mongoengine.GenericReferenceField:

    :param choices: List of Mongoengine document class (or class name) allowed

    .. note:: Without `choices` param, this field allow to reference to
        any document in the application which can be a security issue.
    """

    def __init__(self, *args, **kwargs):
        self.document_class_choices = []
        choices = kwargs.pop("choices", None)
        if choices:
            # Temporary fix for https://github.com/MongoEngine/mongoengine/pull/1060
            for choice in choices:
                if hasattr(choice, "_class_name"):
                    self.document_class_choices.append(choice._class_name)
                else:
                    self.document_class_choices.append(choice)
        super(GenericReference, self).__init__(*args, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        # To deserialize a generic reference, we need a _cls field in addition
        # with the id field
        if not isinstance(value, dict) or not value.get("id") or not value.get("_cls"):
            raise ValidationError("Need a dict with 'id' and '_cls' fields")
        doc_id = value["id"]
        doc_cls_name = value["_cls"]
        if (
            self.document_class_choices
            and doc_cls_name not in self.document_class_choices
        ):
            raise ValidationError(
                "Invalid _cls field `%s`, must be one of %s"
                % (doc_cls_name, self.document_class_choices)
            )
        try:
            doc_cls = get_document(doc_cls_name)
        except NotRegistered:
            raise ValidationError("Invalid _cls field `%s`" % doc_cls_name)
        try:
            doc = doc_cls.objects.get(pk=doc_id)
        except (doc_cls.DoesNotExist, MongoValidationError, ValueError, TypeError):
            raise ValidationError("unknown document %s `%s`" % (doc_cls_name, value))
        return doc

    def _serialize(self, value, attr, obj, **kwargs):
        # Only return the pk of the document for serialization
        if value is None:
            return missing
        return str(value.pk) if isinstance(value.pk, bson.ObjectId) else value.pk


class GenericEmbeddedDocument(fields.Field):

    """
    Dynamic embedded document
    """

    def _deserialize(self, value, attr, data, **kwargs):
        # Cannot deserialize given we have no way knowing wich kind of
        # document is given...
        return missing

    def _serialize(self, value, attr, obj, **kwargs):
        # Create the schema at serialize time to be dynamic
        from ..marshmallow_mongoengine.schema import ModelSchema

        class NestedSchema(ModelSchema):
            class Meta:
                model = type(value)

        return NestedSchema().dump(value)


class Map(fields.Field):

    """
    Marshmallow custom field to map with :class Mongoengine.Map:
    """

    def __init__(self, mapped, **kwargs):
        self.mapped = mapped
        self.schema = getattr(mapped, "schema", None)
        super(Map, self).__init__(**kwargs)

    def _schema_process(self, action, value):
        func = getattr(self.schema, action)
        total = {}
        for k, v in value.items():
            total[k] = func(v)
        return total

    def _serialize(self, value, attr, obj, **kwargs):
        if self.schema:
            return self._schema_process("dump", value)
        else:
            return value

    def _deserialize(self, value, attr, data, **kwargs):
        if self.schema:
            return self._schema_process("load", value)
        else:
            return value


class Skip(fields.Field):

    """
    Marshmallow custom field that just ignore the current field
    """

    def _deserialize(self, value, attr, data, **kwargs):
        return missing

    def _serialize(self, value, attr, obj, **kwargs):
        return missing



"""Test the :mod:`gator.core.data.utils.hash` module."""
import base64
from typing import Any

import pytest

from gator.core.data.utils.hash import make_hash_sha256, make_hashable


class TestMakeHashSha256:
    """Test the :meth:`gator.core.data.utils.hash.make_hash_sha256` function."""

    @pytest.mark.parametrize('x', [1, 1.0, 'hello world', b'test', [], (), {}])
    def test_b64_encoded(self, x: Any) -> None:
        """Test that the hash is base64 encoded."""
        h = make_hash_sha256(x)
        # Decoding and re-encoding a base64 string should result in the same string
        assert h == base64.b64encode(base64.b64decode(h)).decode()

    @pytest.mark.parametrize('x', [1, 1.0, 'hello world', b'test', [], (), {}])
    def test_same_input_same_output(self, x: Any) -> None:
        """Test that the same input always results in the same output."""
        assert make_hash_sha256(x) == make_hash_sha256(x)

    def test_dict_order_agnostic(self) -> None:
        """Test that the hash is order agnostic."""
        d1 = {'a': 1, 'b': 2, 'c': 3}
        d2 = {'c': 3, 'b': 2, 'a': 1}
        assert make_hash_sha256(d1) == make_hash_sha256(d2)


class TestMakeHashable:
    """Test the :meth:`gator.core.data.utils.hash.make_hashable` function."""

    def test_list(self) -> None:
        """Test that a list is converted to a tuple."""
        assert make_hashable([1, 2, 3]) == (1, 2, 3)

    def test_dict(self) -> None:
        """Test that a dict is converted to a tuple of tuples."""
        assert make_hashable({'a': 1, 'b': 2, 'c': 3}) == (('a', 1), ('b', 2), ('c', 3))

    def test_dict_sorted(self) -> None:
        """Test that a dict is sorted by key."""
        assert make_hashable({'c': 3, 'b': 2, 'a': 1}) == (('a', 1), ('b', 2), ('c', 3))

    def test_set(self) -> None:
        """Test that a set is converted to a tuple."""
        assert make_hashable({1, 2, 3}) == (1, 2, 3)

    def test_set_sorted(self) -> None:
        """Test that a set is sorted."""
        assert make_hashable({3, 2, 1}) == (1, 2, 3)

    @pytest.mark.parametrize('x', [1, 1.0, 'hello world', b'test'])
    def test_unchanged(self, x: Any) -> None:
        """Test that an object that is already hashable is returned unchanged."""
        assert make_hashable(x) == x

    def test_recursive(self) -> None:
        """Test that a nested structure is converted recursively."""
        assert make_hashable({
            'a': [1, 2, 3],
            'b': {3, 2, 1},
            'c': {'a': 1, 'b': 2, 'c': {'zyx', 'wvu', 'tsr'}}
        }) == (
            ('a', (1, 2, 3)),
            ('b', (1, 2, 3)),
            ('c', (('a', 1), ('b', 2), ('c', ('tsr', 'wvu', 'zyx'))))
        )



"""Test the :mod:`gator.core.models.institution` module."""
from typing import Optional

import pytest

import gator.core.models.institution as institution


class TestSubInstitutionsList:
    """Test the :class:`gator.core.models.institution._SubInstitutionsList` class."""

    @classmethod
    @pytest.fixture
    def parent_inst(cls) -> institution.Institution:
        """Return a parent institution."""
        return institution.Institution(code='p', name='p', type='uni')

    @classmethod
    @pytest.fixture
    def child_insts(cls) -> list[institution.Institution]:
        """Return a list of child institutions."""
        return [
            institution.Institution(code=f'c{i}', name=f'c{i}', type='uni')
            for i in range(5)
        ]

    @classmethod
    @pytest.fixture
    def child_inst(cls) -> institution.Institution:
        """Return a child institution."""
        return institution.Institution(code='c', name='c', type='uni')

    def test_append(self, parent_inst: institution.Institution,
                    child_inst: institution.Institution) -> None:
        """Test appending to the list.

        Expect that the given `child_inst` is added to the sub-institutions
        list of `parent_inst` and that the parent of `child_inst` is set to
        `parent_inst`.
        """
        sub_inst_list = institution._SubInstitutionsList(parent_inst)
        sub_inst_list.append(child_inst)

        assert child_inst in parent_inst.sub_institutions
        assert child_inst.parent == parent_inst

    def test_extend(self, parent_inst: institution.Institution,
                    child_insts: list[institution.Institution]) -> None:
        """Test extending the list.

        Expect that the given `child_insts` are added to the sub-institutions
        list of `parent_inst` and that the parent of each `child_inst` is set to
        `parent_inst`.
        """
        sub_inst_list = institution._SubInstitutionsList(parent_inst)
        sub_inst_list.extend(child_insts)

        for child_inst in child_insts:
            assert child_inst in parent_inst.sub_institutions
            assert child_inst.parent == parent_inst

    def test_insert(self, parent_inst: institution.Institution,
                    child_inst: institution.Institution) -> None:
        """Test inserting into the list.

        Expect that the given `child_inst` is added to the sub-institutions
        list of `parent_inst` and that the parent of `child_inst` is set to
        `parent_inst`.
        """
        sub_inst_list = institution._SubInstitutionsList(parent_inst)
        sub_inst_list.insert(0, child_inst)

        assert child_inst in parent_inst.sub_institutions
        assert child_inst.parent == parent_inst

    def test_remove(self, parent_inst: institution.Institution,
                    child_inst: institution.Institution) -> None:
        """Test removing from the list.

        Expect that the given `child_inst` is removed from the sub-institutions
        list of `parent_inst` and that the parent of `child_inst` is set to
        None.
        """
        sub_inst_list = institution._SubInstitutionsList(parent_inst)
        sub_inst_list.append(child_inst)
        sub_inst_list.remove(child_inst)

        assert child_inst not in parent_inst.sub_institutions
        assert child_inst.parent is None

    def test_pop(self, parent_inst: institution.Institution,
                 child_insts: list[institution.Institution]) -> None:
        """Test popping from the list.

        Expect that the given `child_inst` is removed from the sub-institutions
        list of `parent_inst` and that the parent of `child_inst` is set to
        None.
        """
        sub_inst_list = institution._SubInstitutionsList(parent_inst)
        sub_inst_list.extend(child_insts)

        for i in range(len(child_insts) - 1, 0, -1):
            # Remove the last child (child_insts[i])
            sub_inst_list.pop()
            assert child_insts[i] not in parent_inst.sub_institutions
            assert child_insts[i].parent is None

            # Check that the remaining children are still in the list
            # and have the correct parent
            for j in range(i):
                assert child_insts[j] in parent_inst.sub_institutions
                assert child_insts[j].parent == parent_inst


class TestInstitution:
    """Test the :class:`gator.core.models.institution.Institution` class."""

    @classmethod
    @pytest.fixture
    def inst(cls) -> institution.Institution:
        """Return an institution with a nested hierarchy."""
        return institution.Institution(
            code='a1',
            name='a1',
            type='root',
            sub_institutions=[
                institution.Institution(
                    code='b1',
                    name='b1',
                    type='child-level-1',
                ),
                institution.Institution(
                    code='b2',
                    name='b2',
                    type='child-level-1',
                    sub_institutions=[
                        institution.Institution(
                            code='c1',
                            name='c1',
                            type='child-level-2',
                        )
                    ]
                ),
                institution.Institution(
                    code='b3',
                    name='b3',
                    type='child-level-1',
                    sub_institutions=[
                        institution.Institution(
                            code='c2',
                            name='c2',
                            type='child-level-2',
                        )
                    ]
                )
            ]
        )

    def test_init(self, inst: institution.Institution) -> None:
        """Test the initialization of an institution.

        Expect that the institution is initialized with the given code,
        name, type, and sub-institutions. Each sub-institution should have
        the proper parent set.
        """
        def _check_sub_insts(curr_inst: institution.Institution) -> None:
            """Recursively check the sub-institutions."""
            for sub_inst in curr_inst.sub_institutions:
                assert sub_inst.parent == curr_inst
                _check_sub_insts(sub_inst)

        _check_sub_insts(inst)

    def test_set_parent(self, inst: institution.Institution) -> None:
        """Test setting the parent of an institution.

        Expect that the parent of the institution is set to the given
        parent institution.
        """
        new_parent_inst = institution.Institution(
            code='a2', name='a2', type='root')
        assert inst.parent is None

        inst.parent = new_parent_inst
        assert inst.parent == new_parent_inst
        assert inst in new_parent_inst.sub_institutions

    def test_set_sub_institutions(self, inst: institution.Institution) -> None:
        """Test setting the sub-institutions of an institution.

        Expect that the sub-institutions of the institution are set to the
        given sub-institutions and that the parent of each sub-institution
        is set to the institution.
        """
        new_sub_insts = [
            institution.Institution(code='b4', name='b4', type='child-level-1'),
            institution.Institution(code='b5', name='b5', type='child-level-1'),
        ]
        assert inst.sub_institutions != new_sub_insts

        inst.sub_institutions = new_sub_insts
        assert inst.sub_institutions == new_sub_insts
        for sub_inst in inst.sub_institutions:
            assert sub_inst.parent == inst

    @pytest.mark.parametrize('type, expected', [
        ('root', ['a1']),
        ('child-level-1', ['b1', 'b2', 'b3']),
        ('child-level-2', ['c1', 'c2']),
        ('child-level-3', [])
    ])
    def test_find_children(self, inst: institution.Institution,
                           type: str, expected: list[str]) -> None:
        """Test finding the children of an institution of a given type.

        Expect that the children of the institution are returned.
        """
        children = inst.find_children(type)
        assert len(children) == len(expected)
        assert {child.code for child in children} == set(expected)

    @pytest.mark.parametrize('start_type, parent_type, expected', [
        ('root', 'root', None),
        ('child-level-1', 'root', 'a1'),
        ('child-level-2', 'root', 'a1'),
        ('child-level-2', 'child-level-1', 'b2')
    ])
    def test_find_parent(self, inst: institution.Institution,
                         start_type: str, parent_type: str,
                         expected: list[str]) -> None:
        """Test finding the parent of an institution of a given type.

        Expect that the parent of the institution is returned.
        """
        start_inst = inst.find_children(start_type)[0]
        parent_inst = start_inst.find_parent(parent_type)

        if parent_inst is None:
            assert expected is None
        else:
            assert parent_inst.code == expected



# type: ignore
"""Base classes for datasets."""
import fnmatch
from abc import ABC, abstractclassmethod, abstractmethod, abstractproperty
from typing import Any, Iterator, Optional, Type, Union

from mongoengine import Document

from gator.core.models.timetable import Session


class Dataset(ABC):
    """A dataset that converts free-form data into mongoengine documents.

    All subclasses should implement functionality for a) pulling data from a
    source (e.g. a database, a file, or an API) and returning it as a list of
    `(id, data)` records via the `get` method, and b) processing each record
    into a :class:`mongoengine.Document` object via the `process` method.
    The `slug`, `name`, and `description` properties should also be implemented
    to provide metadata about the dataset.
    """

    @abstractproperty
    def slug(self) -> str:
        """Return the slug of this dataset."""
        raise NotImplementedError()

    @abstractproperty
    def name(self) -> str:
        """Return the name of this dataset."""
        raise NotImplementedError()

    @abstractproperty
    def description(self) -> str:
        """Return the description of this dataset."""
        raise NotImplementedError()

    @abstractmethod
    def get(self) -> Iterator[tuple[str, Any]]:
        """Return an iterator that lazily yields `(id, data)` records.

        The `id` should be a unique identifier for the record, and the `data`
        can be any hashable object. A hash of the `data` object will be compared
        with a hash stored in the database for the record with the given `id`.
        """
        raise NotImplementedError()

    @abstractmethod
    def process(self, id: str, data: Any) -> Document:
        """Process the given record into a :class:`mongoengine.Document`.

        Args:
            id: The unique identifier for the record.
            data: The data for the record.
        """
        raise NotImplementedError()


class SessionalDataset(Dataset):
    """A dataset that is specific to one or more sessions.

    For example, a dataset of courses offered in a set of particular sessions,
    or a dataset of clubs that are active in a single session.

    All subclasses must implement the `_get_latest_sessions` method to return
    the most up-to-date sessions for this dataset.

    Instance Attributes:
        sessions: The sessions that this dataset is specific to.
    """

    sessions: list[Session]

    def __init__(self,
                 sessions: Optional[list[Union[str, Session]]] = None,
                 session: Optional[Union[str, Session]] = None) -> None:
        """Initialize a SessionalDataset.

        Args:
            sessions: The sessions that this dataset tracks. Each element can
                either be a session code, which will be parsed into a Session
                object, or a Session object itself. If no sessions are provided,
                the most up-to-date sessions will be used.
            session: Instead of providing a list of sessions, a single session
                can be provided. This is equivalent to providing a list with a
                single session. If both `sessions` and `session` are provided,
                `sessions` will take precedence.
        """
        super().__init__()
        # Prefer the `sessions` argument. In the case it is None, use the
        # `session` argument if it is provided, otherwise default to None.
        sessions = sessions or ([session] if session is not None else None)
        if sessions is None:
            # Use the most up-to-date sessions
            self.sessions = self._get_latest_sessions()
        else:
            # Parse the sessions
            self.sessions = [
                Session.from_code(session) if not isinstance(session, Session) else session
                for session in sessions]

    @abstractclassmethod
    def _get_latest_sessions(cls) -> list[Session]:
        """Return the most up-to-date sessions for this dataset.

        Raise a ValueError if the session could not be found.
        """
        raise NotImplementedError()


class Registry:
    """A registry tracking one or more datasets.

    Each dataset is identified by a unique slug. The registry can be used to
    retrieve datasets by their slug, or to filter datasets by a pattern.
    """

    # Private Instance Attributes:
    #     _datasets: A list of datasets to track.
    _datasets: list[Dataset]

    def __init__(self, datasets: Optional[list[Dataset]] = None) -> None:
        """Initialize a Registry.

        Args:
            datasets: A list of datasets to track.
        """
        self._datasets = datasets or []

    def register(self, dataset: Union[Type[Dataset], Dataset], **kwargs: Any) \
            -> None:
        """Register a dataset with the registry.

        Args:
            dataset: A dataset to register. This can either be a class or an
                instance of a dataset. If a class is provided, it will be
                instantiated with the provided keyword arguments.
            **kwargs: Keyword arguments to pass to the dataset initializer
                in the case that `dataset` is a class.
        """
        if isinstance(dataset, type):
            dataset = dataset(**kwargs)
        self._datasets.append(dataset)

    def filter(self, pattern: str) -> Iterator[Dataset]:
        """Filter the datasets by a pattern.

        Args:
            pattern: A Unix-style glob pattern to filter the datasets by.

        Yields:
            Datasets that match the pattern.
        """
        for dataset in self._datasets:
            if fnmatch.fnmatch(dataset.slug, pattern):
                yield dataset

    def all(self) -> Iterator[Dataset]:
        """Return an iterator that yields all datasets."""
        yield from self._datasets



"""Test the :mod:`gator.core.data.utils.serialization` module."""
from typing import Any, Callable

import pytest

from gator.core.data.utils.serialization import nullable_convert, without_keys


class TestNullableConvert:
    """Test the :func:`gator.core.data.utils.serialization.nullable_convert` function."""

    @pytest.mark.parametrize('value, func', [
        (1, int),
        (1.0, float),
        ('hello world', str),
        (b'test', bytes),
        ([1, 2, 3], list),
        ((1, 2, 3), tuple),
        ({'a': 1, 'b': 2, 'c': 3}, dict),
        ({1, 2, 3}, set),
    ])
    def test_identity(self, value: Any, func: Callable) -> None:
        """Test that the function returns the input value."""
        assert nullable_convert(value, func) == value

    @pytest.mark.parametrize('func', [int, float, str, bytes, list, tuple, dict, set])
    def test_handle_none(self, func: Callable) -> None:
        """Test that the function handles None inputs."""
        assert nullable_convert(None, func) is None


class TestWithoutKeys:
    """Test the :func:`gator.core.data.utils.serialization.without_keys` function."""

    def test_empty(self) -> None:
        """Test that the function returns an empty dict when given an empty dict."""
        assert without_keys({}, {'a', 'b'}) == {}

    def test_no_keys(self) -> None:
        """Test that the function returns the input dict when given no keys."""
        d = {'a': 1, 'b': 2, 'c': 3}
        assert without_keys(d, set()) == d

    def test_remove_keys(self) -> None:
        """Test that the function removes the given keys."""
        d = {'a': 1, 'b': 2, 'c': 3}
        assert without_keys(d, {'a', 'c'}) == {'b': 2}

    def test_handles_invalid_keys(self) -> None:
        """Test that the function handles invalid keys."""
        d = {'a': 1, 'b': 2, 'c': 3}
        assert without_keys(d, {'x', 'y'}) == d



# -*- coding: utf-8 -*-


class MarshmallowMongoengineError(Exception):
    """Base exception class from which all exceptions related to
    marshmallow-mongoengine inherit.
    """

    pass


class ModelConversionError(MarshmallowMongoengineError):
    """Raised when an error occurs in converting a Mongoengine construct
    to a marshmallow object.
    """

    pass



# -*- coding: utf-8 -*-
import copy

from mongoengine.base import BaseDocument
import marshmallow as ma
from six import with_metaclass
from ..marshmallow_mongoengine.convert import ModelConverter


DEFAULT_SKIP_VALUES = (None, [], {})


class SchemaOpts(ma.SchemaOpts):
    """Options class for `ModelSchema`.
    Adds the following options:

    - ``model``: The Mongoengine Document model to generate the `Schema`
        from (required).
    - ``model_fields_kwargs``: Dict of {field: kwargs} to provide as
        additionals argument during fields creation.
    - ``model_build_obj``: If true, :Schema load: returns a :model: objects
        instead of a dict (default: True).
    - ``model_converter``: `ModelConverter` class to use for converting the
        Mongoengine Document model to marshmallow fields.
    - ``model_dump_only_pk``: If the document autogenerate it primary_key
        (default behaviour in Mongoengine), ignore it from the incomming data
        (default: False)
    - ``model_skip_values``: Skip the field if it contains one of the given
        values (default: None, [] and {})
    """

    def __init__(self, meta, *args, **kwargs):
        super(SchemaOpts, self).__init__(meta, *args, **kwargs)
        self.model = getattr(meta, "model", None)
        if self.model and not issubclass(self.model, BaseDocument):
            raise ValueError(
                "`model` must be a subclass of mongoengine.base.BaseDocument"
            )
        self.model_fields_kwargs = getattr(meta, "model_fields_kwargs", {})
        self.model_dump_only_pk = getattr(meta, "model_dump_only_pk", False)
        self.model_converter = getattr(meta, "model_converter", ModelConverter)
        self.model_build_obj = getattr(meta, "model_build_obj", True)
        self.model_skip_values = getattr(meta, "model_skip_values", DEFAULT_SKIP_VALUES)


class SchemaMeta(ma.schema.SchemaMeta):
    """Metaclass for `ModelSchema`."""

    # override SchemaMeta
    @classmethod
    def get_declared_fields(mcs, klass, *args, **kwargs):
        """Updates declared fields with fields converted from the
        Mongoengine model passed as the `model` class Meta option.
        """
        declared_fields = kwargs.get("dict_class", dict)()
        # Generate the fields provided through inheritance
        opts = klass.opts
        model = getattr(opts, "model", None)
        if model:
            converter = opts.model_converter()
            declared_fields.update(
                converter.fields_for_model(model, fields=opts.fields)
            )
        # Generate the fields provided in the current class
        base_fields = super(SchemaMeta, mcs).get_declared_fields(klass, *args, **kwargs)
        declared_fields.update(base_fields)
        # Customize fields with provided kwargs
        for field_name, field_kwargs in klass.opts.model_fields_kwargs.items():
            field = declared_fields.get(field_name, None)
            if field:
                # Copy to prevent alteration of a possible parent class's field
                field = copy.copy(field)
                for key, value in field_kwargs.items():
                    setattr(field, key, value)
                declared_fields[field_name] = field
        if opts.model_dump_only_pk and opts.model:
            # If primary key is automatically generated (nominal case), we
            # must make sure this field is read-only
            if opts.model._auto_id_field is True:
                field_name = opts.model._meta["id_field"]
                id_field = declared_fields.get(field_name)
                if id_field:
                    # Copy to prevent alteration of a possible parent class's field
                    id_field = copy.copy(id_field)
                    id_field.dump_only = True
                    declared_fields[field_name] = id_field
        return declared_fields


class ModelSchema(with_metaclass(SchemaMeta, ma.Schema)):
    """Base class for Mongoengine model-based Schemas.

    Example: ::

        from marshmallow_mongoengine import ModelSchema
        from mymodels import User

        class UserSchema(ModelSchema):
            class Meta:
                model = User
    """

    OPTIONS_CLASS = SchemaOpts

    @ma.post_dump
    def _remove_skip_values(self, data, **kwargs):
        to_skip = self.opts.model_skip_values
        return {key: value for key, value in data.items() if value not in to_skip}

    @ma.post_load
    def _make_object(self, data, **kwargs):
        if self.opts.model_build_obj and self.opts.model:
            return self.opts.model(**data)
        else:
            return data

    def update(self, obj, data):
        """Helper function to update an already existing document
        instead of creating a new one.
        :param obj: Mongoengine Document to update
        :param data: incomming payload to deserialize
        :return: an :class UnmarshallResult:

        Example: ::

            from marshmallow_mongoengine import ModelSchema
            from mymodels import User

            class UserSchema(ModelSchema):
                class Meta:
                    model = User

            def update_obj(id, payload):
                user = User.objects(id=id).first()
                result = UserSchema().update(user, payload)
                result.data is user # True

        Note:

            Given the update is done on a existing object, the required param
            on the fields is ignored
        """
        # TODO: find a cleaner way to skip required validation on update
        required_fields = [k for k, f in self.fields.items() if f.required]
        for field in required_fields:
            self.fields[field].required = False
        loaded_data = self._do_load(data, postprocess=False)
        for field in required_fields:
            self.fields[field].required = True
        # Update the given obj fields
        for k, v in loaded_data.items():
            # Skip default values that have been automatically
            # added during unserialization
            if k in data:
                setattr(obj, k, v)
        return obj



# -*- coding: utf-8 -*-
from __future__ import absolute_import

try:
    import marshmallow
except ImportError:
    raise ImportError("marshmallow is not installed.")
else:
    del marshmallow

from ..marshmallow_mongoengine.schema import (
    SchemaOpts,
    ModelSchema,
)

from ..marshmallow_mongoengine.conversion.fields import (
    register_field,
    register_field_builder,
)

from ..marshmallow_mongoengine.convert import (
    ModelConverter,
    fields_for_model,
    convert_field,
    field_for,
)
from ..marshmallow_mongoengine.exceptions import ModelConversionError

__version__ = "0.31.1"
__license__ = "MIT"

__all__ = [
    "ModelSchema",
    "SchemaOpts",
    "ModelConverter",
    "fields_for_model",
    "property2field",
    "column2field",
    "ModelConversionError",
    "convert_field",
    "field_for",
    "fields",
    "register_field_builder",
    "register_field",
]



"""Model definitions for timetable data."""
import math
import re
from functools import lru_cache
from typing import Any, Optional

from mongoengine import Document, EmbeddedDocument, fields

from gator.core.models.common import SerializableEnum
from gator.core.models.institution import Institution, Location


class Session(EmbeddedDocument):
    """A formal division of the academic year.

    The academic year is divided into two seasons: the regular season, which
    runs from September to April, and the summer season, which runs from May
    to August.

    Each season is divided into two subsessions: the first subsession, which
    runs from September to December for the regular season, and from May to
    June for the summer season; and the second subsession, which runs from
    January to April for the regular season, and from July to August for the
    summer season. The whole subsession is the combination of the first and
    second subsessions (e.g. September to April for the regular season and
    May to August for the summer season).

    An offering of a course is associated with a list of sessions, where
    each session corresponds to a specific subsession of a specific season.
    For example, an offering of a course that runs from September to April
    would have two sessions: one for the regular season's first subsession,
    and one for the regular season's second subsession.

    Instance Attributes:
        year: The year in which the season of the session starts. This is the
            year in which the first subsession of the season starts.
        season: The season of the session.
        subsession: The subsession of the session. Can be first, second, or whole.
    """

    # Private Class Attributes:
    #   _SUBSESSION_MAP: A mapping from season name to a mapping from subsession
    #       name to its code (canonical representation).
    #   _INVERSE_SUBSESSION_MAP: The inverse of `_SUBSESSION_MAP`. A mapping
    #       from season name to a mapping from subsession code to its name.
    #   _SUBSESSION_TO_SEASON_MAP: A mapping from subsession code to season
    #       name.
    #   _SEASON_ORDER: A mapping from season name to its sort order.
    #   _SUBSESSION_ORDER: A mapping from subsession name to its sort order.
    _SUBSESSION_MAP: dict[str, dict[str, str]] = {
        'regular': {'first': '9', 'second': '1'},
        'summer': {'first': '5F', 'second': '5S', 'whole': '5'}
    }
    _INVERSE_SUBSESSION_MAP: dict[str, dict[str, str]] = {
        s: {v: k for k, v in d.items()} for s, d in _SUBSESSION_MAP.items()
    }
    _SUBSESSION_TO_SEASON_MAP: dict[str, str] = {
        v: k for k, d in _SUBSESSION_MAP.items() for v in d.values()
    }
    _SEASON_ORDER: dict[str, int] = {
        'regular': 0,
        'summer': 1
    }
    _SUBSESSION_ORDER: dict[str, int] = {
        'first': 0,
        'second': 1,
        'whole': 2
    }

    year: int = fields.IntField(required=True, min_value=0)  # type: ignore
    season: str = fields.StringField(
        required=True, choices=['regular', 'summer'])  # type: ignore
    subsession: str = fields.StringField(
        required=True, choices=['first', 'second', 'whole'])  # type: ignore

    def __init__(self, year: int, season: str, subsession: str, *args: Any, **kwargs: Any) -> None:
        """Initialize a session with the given year, season, and subsession.

        Args:
            year: The year of the session.
            season: The season of the session.
            subsession: The subsession of the session.

        Note:
            The year, season, and subsession are passed as positional arguments
            for convenience, but they are actually stored as keyword arguments
            in the `kwargs` dictionary. If `year`, `season`, or `subsession`
            are passed as keyword arguments, they will be ignored, meaning that
            the positional arguments always take precedence.
        """
        kwargs['year'] = year
        kwargs['season'] = season
        kwargs['subsession'] = subsession
        super().__init__(*args, **kwargs)

    @property
    def code(self) -> str:
        """Return the canonical representation of the session.

        The canonical representation of a session is a string that uniquely
        identifies the subsession for a given season in a given year in
        accordance with the University of Toronto's academic calendar.

        For non-whole subsessions, this representation is a string of the form
        `YYYYX`, where `YYYY` is the year of the session, and `X` is the
        subsession code of the session. The subsession code is a unique
        identifier for the subsession of the session, and is defined as
        follows:
            - `9` for the first (fall) subsession of the regular season.
            - `1` for the second (winter) subsession of the regular season.
            - `5F` for the first subsession of the summer season.
            - `5S` for the second subsession of the summer season.
            - `5` for the whole subsession of the summer season.

        The whole (fall-winter) subsession of the regular season is a special
        case, and is represented as a range of two sessions, one for the first
        subsession (in year :math:`Y`) and one for the second subsession (in
        year :math:`Y+1`), separated by a hyphen. For example, the whole
        subsession of the regular season in 2022 is represented as the string
        `20229-20231`.

        Remarks:
            - Years will be zero-padded up to four digits. For example, the
                year 2022 will be represented as `2022`, and the year 1 will be
                represented as `0001`.

        Examples:
            >>> Session(2022, 'regular', 'first').code  # Fall 2022
            '20229'
            >>> Session(2023, 'regular', 'second').code  # Winter 2023
            '20231'
            >>> Session(2022, 'regular', 'whole').code  # Fall 2022 - Winter 2023
            '20229-20231'
            >>> Session(2023, 'summer', 'first').code  # Summer 2023 (First Subsession)
            '20235F'
            >>> Session(2023, 'summer', 'second').code  # Summer 2023 (Second Subsession)
            '20235S'
            >>> Session(2023, 'summer', 'whole').code  # Summer 2023 (Whole Session)
            '20235'
        """
        if self.season == 'regular' and self.subsession == 'whole':
            # Recursively call the code property on the first and second
            return '-'.join([
                Session(self.year, self.season, 'first').code,
                Session(self.year + 1, self.season, 'second').code
            ])

        year = str(self.year).zfill(4)
        return f'{year}{self._SUBSESSION_MAP[self.season][self.subsession]}'

    @property
    def human_str(self) -> str:
        """A human-readable representation of this session.

        Examples:
            >>> Session(2022, 'regular', 'first').human_str  # Fall 2022
            'Fall 2022'
            >>> Session(2023, 'regular', 'second').human_str  # Winter 2023
            'Winter 2023'
            >>> Session(2022, 'regular', 'whole').human_str  # Fall 2022 - Winter 2023
            'Fall 2022 - Winter 2023'
            >>> Session(2023, 'summer', 'first').human_str  # Summer 2023 (First Subsession)
            'Summer 2023 (First Subsession)'
            >>> Session(2023, 'summer', 'second').human_str  # Summer 2023 (Second Subsession)
            'Summer 2023 (Second Subsession)'
            >>> Session(2023, 'summer', 'whole').human_str  # Summer 2023 (Whole Session)
            'Summer 2023 (Whole Session)'
        """
        if self.season == 'regular' and self.subsession == 'whole':
            # Recursively call the human_str property on the first and second
            return ' - '.join([
                Session(self.year, self.season, 'first').human_str,
                Session(self.year + 1, self.season, 'second').human_str,
            ])

        subsession = {
            'regular': {
                'first': 'Fall %s',
                'second': 'Winter %s'
            },
            'summer': {
                'first': 'Summer %s (First Subsession)',
                'second': 'Summer %s (Second Subsession)',
                'whole': 'Summer %s (Whole Session)',
            },
        }[self.season][self.subsession]
        return subsession % str(self.year).zfill(4)

    def __str__(self) -> str:
        """Return a string representation of the session."""
        return self.code

    def __repr__(self) -> str:
        """Return a string representation of the session."""
        return f'Session({self.year}, {self.season!r}, {self.subsession!r})'

    def __eq__(self, other: 'Session') -> bool:
        """Return whether this session is the same as another session.

        Equality is defined as having the same year, season, and subsession.

        Examples:
            >>> Session(2022, 'regular', 'first') == Session(2022, 'regular', 'first')
            True
            >>> Session(1999, 'summer', 'whole') == Session(1999, 'summer', 'whole')
            True
            >>> Session(2022, 'regular', 'first') == Session(2022, 'regular', 'second')
            False
            >>> Session(2022, 'regular', 'first') == Session(2022, 'summer', 'first')
            False
        """
        return all([
            self.year == other.year,
            self.season == other.season,
            self.subsession == other.subsession,
        ])

    def __ne__(self, other: 'Session') -> bool:
        """Return whether this session is not the same as another session.

        Remarks:
            This is the inverse of the :meth:`__eq__` method. See the
            documentation of that method for more details.
        """
        return not self == other

    def __lt__(self, other: 'Session') -> bool:
        """Return whether this session occurs before the other session.

        A session occurs before another if it is in an earlier year, same year
        but in an earlier season (i.e. regular before summer), or same year and
        season but in an earlier subsession (i.e. first before second or whole).

        Examples:
            >>> Session(2022, 'regular', 'first') < Session(2022, 'regular', 'second')
            True
            >>> Session(2022, 'regular', 'second') < Session(2022, 'regular', 'whole')
            True
            >>> Session(2022, 'regular', 'first') < Session(2022, 'summer', 'first')
            True
            >>> Session(2022, 'regular', 'first') < Session(2023, 'regular', 'first')
            True
            >>> Session(2022, 'regular', 'first') < Session(2022, 'regular', 'first')
            False
        """
        return any([
            # This session is in an EARLIER YEAR
            self.year < other.year,
            # This session is in the SAME YEAR but in an EARLIER SEASON
            all([
                self.year == other.year,
                self._SEASON_ORDER[self.season] < self._SEASON_ORDER[other.season]
            ]),
            # This session is in the SAME YEAR and SEASON but in an EARLIER SUBSESSION
            all([
                self.year == other.year,
                self._SEASON_ORDER[self.season] == self._SEASON_ORDER[other.season],
                self._SUBSESSION_ORDER[self.subsession] < self._SUBSESSION_ORDER[other.subsession]
            ])
        ])

    def __le__(self, other: 'Session') -> bool:
        """Return whether this session occurs before or at the same time as another session."""
        return self < other or self == other

    def __gt__(self, other: 'Session') -> bool:
        """Return whether this session occurs after the other session.

        Remarks:
            This is the inverse of the :meth:`__lt__` method. See the
            documentation of that method for more details.
        """
        return not self < other

    def __ge__(self, other: 'Session') -> bool:
        """Return whether this session occurs after or at the same time as another session."""
        return self > other or self == other

    @classmethod
    def from_code(cls, code: str) -> 'Session':
        """Return a session from its canonical representation.

        Args:
            code: The canonical representation of the session. This is case
                insensitive as the input will be converted to uppercase.

        Raises:
            ValueError: If the given code is invalid or malformed.

        Examples:
            >>> Session.from_code('20229')  # Fall 2022
            Session(2022, 'regular', 'first')
            >>> Session.from_code('20231')  # Winter 2023
            Session(2023, 'regular', 'second')
            >>> Session.from_code('20229-20231')  # Fall 2022 - Winter 2023
            Session(2022, 'regular', 'whole')
            >>> Session.from_code('20235F')  # Summer 2023 (First Subsession)
            Session(2023, 'summer', 'first')
            >>> Session.from_code('20235S')  # Summer 2023 (Second Subsession)
            Session(2023, 'summer', 'second')
            >>> Session.from_code('20235')  # Summer 2023 (Whole Session)
            Session(2023, 'summer', 'whole')
        """
        if '-' in code:
            start, end = (cls.from_code(c) for c in code.split('-'))
            # Ensure that the start and end sessions are consecutive;
            # in the same season; and the subsessions are consecutive
            if start.year != end.year - 1 or start.season != end.season or \
                    start.subsession != 'first' or end.subsession != 'second':
                raise ValueError(f'Invalid session code: {code}')

            return cls(start.year, start.season, 'whole')

        # Validate the code (must be 4 digits followed by a subsession code)
        code = code.upper()
        all_subsession_codes = [code for v in cls._SUBSESSION_MAP.values() for code in v.values()]
        quoted_subsession_codes = [f'"{code}"' for code in all_subsession_codes]
        if not re.match(fr'^\d{{4}}(?:{"|".join(all_subsession_codes)})$', code):
            raise ValueError(
                f'invalid session code: {code}; expected a 4-digit year '
                f'followed by one of the following subsession codes: '
                f'{", ".join(quoted_subsession_codes)}')

        # Get the year from the code
        year = int(code[:4])
        # Get the subsession code from the code
        subsession_code = code[4:]
        # Get the season and subsession from the subsession code
        # These are guaranteed to exist because of the validation above
        season = cls._SUBSESSION_TO_SEASON_MAP[subsession_code]
        subsession = cls._INVERSE_SUBSESSION_MAP[season][subsession_code]
        return cls(year, season, subsession)


class WeeklyRepetitionSchedule(EmbeddedDocument):
    """A schedule for an event that repeats weekly.

    Represented as an integer whose binary representation is
    interpreted as a set of :math:`k` bits, where :math:`k` is the number of
    weeks in a single repeating period. The :math:`i`-th bit is 1 if the
    meeting occurs in the :math:`i`-th week of the repeating period, and 0
    otherwise. For example, if a meeting occurs every week, then :math:`k = 1`
    and the bit is `1`. If a meeting occurs every other week, then
    :math:`k = 2` and the bits are `10`.

    Instance Attributes:
        schedule: The integer representing the repetition schedule.
    """

    schedule: int = fields.IntField(min_value=1, default=1)  # type: ignore

    @property
    def is_alternating(self) -> bool:
        """Whether this meeting does not regularly occur every week.

        A schedule is alternating (i.e. does not occur every week) if there is
        at least one week :math:`i` such that the :math:`i`-th bit is 0.

        Examples:
            >>> WeeklyRepetitionSchedule(schedule=0b1010).is_alternating
            True
            >>> WeeklyRepetitionSchedule(schedule=0b1111).is_alternating
            False
        """
        k = math.ceil(math.log2(self.schedule))
        return self.schedule > 1 and self.schedule != (1 << k) - 1

    @property
    def schedule_bits(self) -> str:
        """The binary representation of the schedule.

        Examples:
            >>> WeeklyRepetitionSchedule(schedule=0b1).schedule_bits
            '1'
            >>> WeeklyRepetitionSchedule(schedule=0b1010).schedule_bits
            '1010'
        """
        return bin(self.schedule)[2:]  # Remove the '0b' prefix

    @property
    def weeks(self) -> list[int]:
        """A list of the weeks in which this meeting occurs.

        Each week is represented as an integer between 1 and the number of
        weeks in a repeating period, inclusive. The weeks are sorted in
        ascending order.

        Examples:
            >>> WeeklyRepetitionSchedule(schedule=0b1).weeks
            [1]
            >>> WeeklyRepetitionSchedule(schedule=0b1010).weeks
            [2, 4]
        """
        return [i + 1 for i, bit in enumerate(reversed(self.schedule_bits)) if bit == '1']


class SectionMeeting(EmbeddedDocument):
    """A single meeting of a section.

    Instance Attributes:
        day: The day of the week that this meeting occurs on, represented as an
            integer, where Monday is 0 and Sunday is 6.
        start_time: The time that this meeting starts, represented as the number
            of milliseconds since midnight (00:00:00).
        end_time: The time that this meeting ends, represented as the number of
            milliseconds since midnight (00:00:00).
        session: The session that this meeting occurs in.
        location: The location of this meeting, or None if this meeting has no
            assigned location.
        repetition_schedule: The repetition schedule of this meeting. Defaults
            to a schedule that occurs every week.
    """

    day: int = fields.IntField(required=True, min_value=0, max_value=6)  # type: ignore
    start_time: int = fields.IntField(required=True, min_value=0, max_value=86400000)  # type: ignore
    end_time: int = fields.IntField(required=True, min_value=0, max_value=86400000)  # type: ignore
    session: Session = fields.EmbeddedDocumentField(Session, required=True)  # type: ignore
    location: Location = fields.EmbeddedDocumentField(Location, null=True, default=None)  # type: ignore
    repetition_schedule: WeeklyRepetitionSchedule = fields.EmbeddedDocumentField(
        WeeklyRepetitionSchedule, default=WeeklyRepetitionSchedule)  # type: ignore


class TeachingMethod(SerializableEnum):
    """A class representing the teaching method for a section."""

    LECTURE = 'LEC'
    TUTORIAL = 'TUT'
    PRACTICAL = 'PRA'


class SectionDeliveryMode(SerializableEnum):
    """A class representing mode of delivery for a section."""

    CLASS = 'CLASS'
    ONLINE_SYNC = 'ONLSYNC'
    ONLINE_ASYNC = 'ONLASYNC'
    IN_PERSON = 'INPER'
    SYNC = 'SYNC'
    ASYNC = 'ASYNC'
    # idk why they keep changing delivery modes
    ASYIF = 'ASYIF'
    SYNIF = 'SYNIF'


class Instructor(EmbeddedDocument):
    """A course instructor.

    Instance Attributes:
        first_name: The first name of this instructor.
        last_name: The last name of this instructor.
    """

    first_name: str = fields.StringField(required=True)  # type: ignore
    last_name: str = fields.StringField(required=True)  # type: ignore


class EnrolmentInfo(EmbeddedDocument):
    """Enrolment information for a section.

    Instance Attributes:
        current_enrolment: The number of students currently enrolled in this
            section.
        max_enrolment: The maximum number of students that can be enrolled in
            this section.
        has_waitlist: Whether this section has a waitlist.
        current_waitlist_size: The number of students currently on the waitlist
            for this section.
        enrolment_indicator: A string representing the enrollment indicator
            for this section, or None if there is no enrollment indicator.
    """

    current_enrolment: Optional[int] = fields.IntField(null=True, min_value=0, default=None)  # type: ignore
    max_enrolment: Optional[int] = fields.IntField(null=True, min_value=0, default=None)  # type: ignore
    has_waitlist: bool = fields.BooleanField(default=False)  # type: ignore
    current_waitlist_size: Optional[int] = fields.IntField(null=True, min_value=0, default=None)  # type: ignore
    enrolment_indicator: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    # TODO: Add enrolment controls


class Section(EmbeddedDocument):
    """A section of a course.

    Instance Attributes:
        teaching_method: The teaching method for this section.
        section_number: The number of this section, represented as a string.
        meetings: A list of meetings available for this section.
        instructors: A list of instructors teaching this section.
        delivery_modes: A list of delivery modes for this section. The i-th
            element of this list corresponds to the delivery mode in the
            i-th term that the course runs in.
        subtitle: The subtitle of this section, or None if there is no subtitle.
            Among other things, this is used to distinguish between the content
            matter when a course offers curriculum on multiple topics. For
            example, an introductory algorithms course might have a section
            that uses Python and another section that uses Java.
        cancelled: Whether this section is cancelled.
        enrolment_info: The enrolment information for this section.
        notes: A list of HTML strings.
        linked_sections: A list of sections that are linked to this section.
            These are strings of the form 'LEC 001' or 'TUT 001' (i.e. teaching
            method and section number).
    """

    teaching_method: TeachingMethod = fields.EnumField(TeachingMethod, required=True)  # type: ignore
    section_number: str = fields.StringField(required=True)  # type: ignore
    meetings: list[SectionMeeting] = fields.EmbeddedDocumentListField('SectionMeeting', required=True)  # type: ignore
    instructors: list[Instructor] = fields.EmbeddedDocumentListField('Instructor', required=True)  # type: ignore
    delivery_modes: list[SectionDeliveryMode] = fields.ListField(
        fields.EnumField(SectionDeliveryMode), required=True)  # type: ignore
    subtitle: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    cancelled: bool = fields.BooleanField(default=False)  # type: ignore
    enrolment_info: EnrolmentInfo = fields.EmbeddedDocumentField(EnrolmentInfo, default=EnrolmentInfo)  # type: ignore
    notes: list[str] = fields.ListField(fields.StringField(), default=list)  # type: ignore
    # TODO: Section dependencies. This should be represented as a graph, where
    # each node is a section and each edge is a dependency (which may or may not
    # be bidirectional).
    linked_sections: list[str] = fields.ListField(fields.StringField(), default=list)  # type: ignore

    @property
    def code(self) -> str:
        """Return a string representing the code of this section.

        This is a combination of the teaching method and the section
        number, separated by a hyphen.

        If the teaching method is None, then the section number is returned
        instead.
        """
        if self.teaching_method is None:
            return self.section_number
        else:
            return f'{self.teaching_method.value}-{self.section_number}'


class Term(SerializableEnum):
    """The term in which a course is offered.

    Class Attributes:
        FIRST_SEMESTER: The first semester of the academic year.
        SECOND_SEMESTER: The second semester of the academic year.
        FULL_YEAR: The full academic year. Spans both semesters.
    """

    FIRST_SEMESTER = 'F'
    SECOND_SEMESTER = 'S'
    FULL_YEAR = 'Y'


class InstructionLevel(SerializableEnum):
    """The level of instruction for a course."""

    UNDERGRADUATE = 'undergraduate'


class CategoricalRequirement(EmbeddedDocument):
    """A breadth or distribution requirement.

    Instance Attributes:
        code: A unique code for this categorical requirement.
        type: The type of this categorical requirement. Either 'breadth' or
            'distribution'.
        name: The name of this categorical requirement.
        description: A description of this categorical requirement.
        institutions: A list of institutions where this categorical requirement
            applies. For example, a breadth requirement might only apply to
            students in the Faculty of Arts and Science, while a distribution
            requirement might apply to all students. There should be at least
            one institution in this list.
    """

    code: str = fields.StringField(unique=True, primary_key=True)  # type: ignore
    type: str = fields.StringField(required=True, choices=['breadth', 'distribution'])  # type: ignore
    name: str = fields.StringField()  # type: ignore
    description: str = fields.StringField()  # type: ignore
    institutions: list[Institution] = fields.ListField(
        fields.ReferenceField(Institution), required=True)  # type: ignore


class Course(Document):
    """An instance of a course offered at UofT for a specific term.

    Instance Attributes:
        id: A unique identifier for this course.
        code: The code of this course.
        name: The name of this course.
        sections: A list of sections available for this course.
        sessions: A list of sessions in which this course is offered.
        term: The term this course is offered in.
        credits: The number of credits this course is worth.
        institution: The institution that offers this course. This is a sub-tree
            of the institution hierarchy.
        title: The title of this course. This is usually the same as the name.
        instruction_level: The level of instruction for this course.
            For example, undergraduate or graduate.
        categorical_requirements: Breadth and distribution requirements that this
            course fulfills.
        description: The description of this course.
        prerequisites: Prerequisties for this course.
        corequisites: Corequisites for this course.
        exclusions: Exclusions for this course.
        recommended_preparation: Recommended preparations to complete before this course.
        cancelled: Whether this course is cancelled.
        tags: Tags for this course.
        notes: A list of HTML strings.
    """

    id: str = fields.StringField(primary_key=True)  # type: ignore
    code: str = fields.StringField()  # type: ignore
    name: str = fields.StringField()  # type: ignore
    sections: list[Section] = fields.EmbeddedDocumentListField('Section')  # type: ignore
    sessions: list[Session] = fields.EmbeddedDocumentListField('Session')  # type: ignore
    term: Term = fields.EnumField(Term)  # type: ignore
    credits: float = fields.FloatField()  # type: ignore
    institution: Institution = fields.ReferenceField(Institution)  # type: ignore
    # Metadata fields
    title: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    instruction_level: Optional[InstructionLevel] = fields.EnumField(
        InstructionLevel, null=True, default=None)  # type: ignore
    description: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    categorical_requirements: list[CategoricalRequirement] = fields.EmbeddedDocumentListField(
        CategoricalRequirement, default=list)  # type: ignore
    prerequisites: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    corequisites: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    exclusions: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    recommended_preparation: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    cancelled: bool = fields.BooleanField(default=False)  # type: ignore
    tags: list[str] = fields.ListField(fields.StringField(), default=list)  # type: ignore
    notes: list[str] = fields.ListField(fields.StringField(), default=list)  # type: ignore

    @property
    @lru_cache
    def section_codes(self) -> set[str]:
        """Return a set of section codes for this course."""
        return {section.code for section in self.sections}

    @property
    @lru_cache
    def level(self) -> int:
        """Return the level of this course.

        If the course is not a level course, return 0.
        """
        m = re.search(r'(?:[^\d]*)(\d+)', self.code)
        if not m:
            return 0
        else:
            return int(math.floor(int(m.group(1)) / 100.0)) * 100



from marshmallow import validate


def _get_base_field_kwargs():
    return {"validate": []}


class MetaParam(object):
    def __init__(self):
        self.field_kwargs = _get_base_field_kwargs()

    def apply(self, field_kwargs=None):
        if not field_kwargs:
            field_kwargs = _get_base_field_kwargs()
        for key, value in self.field_kwargs.items():
            if key == "validate":
                if "validate" not in field_kwargs:
                    field_kwargs["validate"] = []
                field_kwargs["validate"] += value
            else:
                field_kwargs[key] = value
        return field_kwargs


class RequiredParam(MetaParam):
    def __init__(self, field_me):
        super(RequiredParam, self).__init__()
        required = getattr(field_me, "required")
        # If the field has a default value, we don't have to enforce the
        # require check
        if required and getattr(field_me, "default", None) is None:
            self.field_kwargs["required"] = required


class LengthParam(MetaParam):
    def __init__(self, field_me):
        super(LengthParam, self).__init__()
        # Add a length validator for max_length/min_length
        maxmin_args = {}
        if hasattr(field_me, "max_length"):
            maxmin_args["max"] = field_me.max_length
        if hasattr(field_me, "min_length"):
            maxmin_args["min"] = field_me.min_length
        self.field_kwargs["validate"].append(validate.Length(**maxmin_args))


class RegexParam(MetaParam):
    def __init__(self, field_me):
        super(RegexParam, self).__init__()
        regex = getattr(field_me, "regex", None)
        if regex:
            self.field_kwargs["validate"].append(validate.Regexp(regex))


class SizeParam(MetaParam):
    def __init__(self, field_me):
        super(SizeParam, self).__init__()
        # Add a length validator for max_length/min_length
        maxmin_args = {}
        if hasattr(field_me, "max_value"):
            maxmin_args["max"] = field_me.max_value
        if hasattr(field_me, "min_value"):
            maxmin_args["min"] = field_me.min_value
        self.field_kwargs["validate"].append(validate.Range(**maxmin_args))


class DescriptionParam(MetaParam):
    def __init__(self, field_me):
        super(DescriptionParam, self).__init__()
        description = getattr(field_me, "help_text", None)
        if description:
            self.field_kwargs.setdefault("metadata", {})["description"] = description


class AllowNoneParam(MetaParam):
    def __init__(self, field_me):
        super(AllowNoneParam, self).__init__()
        allow_none = getattr(field_me, "null", None)
        if allow_none:
            self.field_kwargs["allow_none"] = True


class ChoiceParam(MetaParam):
    def __init__(self, field_me):
        super(ChoiceParam, self).__init__()
        choices = getattr(field_me, "choices", None)
        labels = None
        if choices and isinstance(choices[0], (list, tuple)):
            choices, labels = zip(*choices)
        if choices:
            self.field_kwargs["validate"].append(validate.OneOf(choices, labels))


class PrecisionParam(MetaParam):
    def __init__(self, field_me):
        super(PrecisionParam, self).__init__()
        precision = getattr(field_me, "precision", None)
        if precision:
            self.field_kwargs["places"] = precision



"""Test the :mod:`gator.core.models.timetable` module."""
import pytest

import gator.core.models.timetable as timetable


class TestSession:
    """Test the :class:`gator.core.models.timetable.Session` class."""

    # Private Class Attributes
    #   _ALL_YEARS: A list of all possible years for a session for the
    #       `__eq__` and `__ne__` tests.
    #   _ALL_SEASONS: A list of all possible seasons for a session for the
    #       `__eq__` and `__ne__` tests.
    #   _ALL_SUBSESSIONS: A list of all possible subsessions for a session for
    #       the `__eq__` and `__ne__` tests.
    _ALL_YEARS = [1996, 1997, 37]
    _ALL_SEASONS = ['regular', 'summer']
    _ALL_SUBSESSIONS = ['first', 'second', 'whole']

    @pytest.mark.parametrize('session, expected_code', [
        (timetable.Session(1996, 'regular', 'first'), '19969'),
        (timetable.Session(1997, 'regular', 'second'), '19971'),
        (timetable.Session(1996, 'regular', 'whole'), '19969-19971'),
        (timetable.Session(1997, 'summer', 'first'), '19975F'),
        (timetable.Session(1997, 'summer', 'second'), '19975S'),
        (timetable.Session(1997, 'summer', 'whole'), '19975'),
        (timetable.Session(37, 'regular', 'first'), '00379')
    ])
    def test_code(self, session: timetable.Session, expected_code: str) -> None:
        """Test the :attr:`gator.core.models.timetable.Session.code` property."""
        assert session.code == expected_code

    @pytest.mark.parametrize('session, expected_human_str', [
        (timetable.Session(1996, 'regular', 'first'), 'Fall 1996'),
        (timetable.Session(1997, 'regular', 'second'), 'Winter 1997'),
        (timetable.Session(1996, 'regular', 'whole'), 'Fall 1996 - Winter 1997'),
        (timetable.Session(1997, 'summer', 'first'), 'Summer 1997 (First Subsession)'),
        (timetable.Session(1997, 'summer', 'second'), 'Summer 1997 (Second Subsession)'),
        (timetable.Session(1997, 'summer', 'whole'), 'Summer 1997 (Whole Session)'),
        (timetable.Session(37, 'regular', 'first'), 'Fall 0037')
    ])
    def test_human_str(self, session: timetable.Session,
                       expected_human_str: str) -> None:
        """Test the :attr:`gator.core.models.timetable.Session.human_str` property."""
        assert session.human_str == expected_human_str

    def test_eq(self) -> None:
        """Test the :meth:`gator.core.models.timetable.Session.__eq__` method."""
        for year in self._ALL_YEARS:
            for season in self._ALL_SEASONS:
                for subsession in self._ALL_SUBSESSIONS:
                    session = timetable.Session(year, season, subsession)
                    # Test equality with itself
                    assert session == session
                    # Test equality with a copy
                    assert session == timetable.Session(year, season, subsession)

    def test_ne(self) -> None:
        """Test the :meth:`gator.core.models.timetable.Session.__ne__` method."""
        for year in self._ALL_YEARS:
            for season in self._ALL_SEASONS:
                for subsession in self._ALL_SUBSESSIONS:
                    session = timetable.Session(year, season, subsession)
                    # Test inequality with a different year
                    assert session != timetable.Session(year + 1, season, subsession)
                    # Test inequality with a different season
                    season_not_equal = 'summer' if season == 'regular' else 'regular'
                    assert session != timetable.Session(
                        year, season_not_equal, subsession)
                    # Test inequality with a different subsession
                    subsession_not_equal = 'second' if subsession == 'first' else 'first'
                    assert session != timetable.Session(
                        year, season, subsession_not_equal)

    def test_lt(self) -> None:
        """Test the :meth:`gator.core.models.timetable.Session.__lt__` method."""
        # Test that a session is less than another session with a greater year
        for season in self._ALL_SEASONS:
            for subsession in self._ALL_SUBSESSIONS:
                assert timetable.Session(1996, season, subsession) < \
                    timetable.Session(1997, season, subsession)
        # Test that a session is less than another session with a greater season
        for subsession in self._ALL_SUBSESSIONS:
            assert timetable.Session(1997, 'regular', subsession) < \
                timetable.Session(1997, 'summer', subsession)
        # Test that a session is less than another session with a greater subsession
        assert timetable.Session(1997, 'regular', 'first') < \
            timetable.Session(1997, 'regular', 'second') < \
            timetable.Session(1997, 'regular', 'whole')
        assert timetable.Session(1997, 'summer', 'first') < \
            timetable.Session(1997, 'summer', 'second') < \
            timetable.Session(1997, 'summer', 'whole')

    def test_sort(self) -> None:
        """Test the :meth:`gator.core.models.timetable.Session.sort` method."""
        sessions = [
            timetable.Session(1996, 'regular', 'first'),
            timetable.Session(1997, 'regular', 'second'),
            timetable.Session(1996, 'regular', 'whole'),
            timetable.Session(1997, 'summer', 'first'),
            timetable.Session(2023, 'regular', 'first'),
            timetable.Session(1997, 'summer', 'second'),
            timetable.Session(1997, 'summer', 'whole'),
            timetable.Session(37, 'regular', 'first')
        ]
        assert sorted(sessions) == [
            timetable.Session(37, 'regular', 'first'),
            timetable.Session(1996, 'regular', 'first'),
            timetable.Session(1996, 'regular', 'whole'),
            timetable.Session(1997, 'regular', 'second'),
            timetable.Session(1997, 'summer', 'first'),
            timetable.Session(1997, 'summer', 'second'),
            timetable.Session(1997, 'summer', 'whole'),
            timetable.Session(2023, 'regular', 'first')
        ]



# coding=utf-8

import sys

if sys.version_info[:2] >= (3, 8):
    from collections.abc import Sequence, MutableMapping
else:
    from collections import Sequence, MutableMapping

import copy

__version__ = '0.5.0'


class Hook(object):
    def _before_add(self, key, item):
        """
        before add a item to the object will call this method.

        example: obj[key] = item
        """

    def _after_add(self, key, item):
        """
        like _before_add, but after add.
        """

    def _before_del(self, key, item):
        """
        before delete a item from the object will call this method.

        example: del obj[key]
        """

    def _after_del(self, key, item):
        """
        like _before_del, but after del.
        """


class Check:
    def _add_check(self, key, value):
        pass

    def _del_check(self, key, value):
        pass


class Garud:

    def _add_garud(self, key, value):
        return key, value

    def _del_garud(self, key):
        return key


class List(Hook, Sequence):
    """
    list like.
    """

    def __init__(self, initlist=None, hook_when_init=True):
        """

        :param initlist: iterable object
        :param hook_when_init: run hook points when it is True
        """

        self._data = []

        if initlist:
            if hook_when_init:
                self.extend(initlist)
            else:
                self._data.extend(initlist)

    def __setitem__(self, i, item):  # x[i] = item, del and add

        if isinstance(i, slice):
            if not hasattr(item, '__contains__'):
                raise TypeError('can only assign an iterable')

            start, stop, step = i.indices(len(self))

            ########################################################
            if step == 1:
                for one in range(stop - start):
                    del self[start]

                _i = start
                for one in item:
                    self.insert(_i, one)
                    _i += 1

            else:

                if step > 1:
                    slice_size = (stop - start) // step

                    if 0 < (stop - start) % step < step:
                        slice_size += 1
                else:
                    slice_size = (start - stop) // abs(step)

                    if 0 < (start - stop) % abs(step) < abs(step):
                        slice_size += 1

                slice_size = 0 if slice_size < 0 else slice_size
                if slice_size != len(item):
                    raise ValueError('attempt to assign sequence of size {} to extended slice of size {}'.format(
                        len(item), slice_size
                    ))

                _i = start
                for one in item:
                    self[_i] = one
                    _i += step

        else:
            del self[i]
            self.insert(i, item)

    # all del action will be here
    def __delitem__(self, i):  # del x[i], del
        item = self[i]
        self._before_del(key=i, item=item)
        del self._data[i]
        self._after_del(key=i, item=item)

    def append(self, item):  # add
        self.insert(len(self), item)

    # all add action will be here
    def insert(self, i, item):
        self._before_add(key=i, item=item)
        self._data.insert(i, item)
        self._after_add(key=i, item=item)

    def pop(self, i=-1):  # del
        x = self[i]
        del self[i]
        return x

    def remove(self, item):  # del
        i = self.index(item)
        del self[i]

    def clear(self):  # del
        for i in range(len(self)):
            self.pop()

    def extend(self, other):  # add
        for item in other:
            self.append(item)

    def __iadd__(self, other):  # x += y, add
        self.extend(other)
        return self

    def __imul__(self, n):  # x *= y, add
        if not isinstance(n, int):
            raise TypeError("can't multiply sequence by non-int of type '{}'".format(type(n)))

        if n <= 0:
            self.clear()

        if n == 1:
            pass

        elif n > 1:
            old_data = copy.copy(self._data)
            for time in range(n - 1):
                self.extend(old_data)

        return self

    def __getitem__(self, i): return self._data[i]

    def __len__(self): return len(self._data)

    def __repr__(self): return repr(self._data)

    def __lt__(self, other): return self._data < self.__cast(other)

    def __le__(self, other): return self._data <= self.__cast(other)

    def __eq__(self, other): return self._data == self.__cast(other)

    def __gt__(self, other): return self._data > self.__cast(other)

    def __ge__(self, other): return self._data >= self.__cast(other)

    def __cast(self, other):
        return other._data if isinstance(other, List) else other

    def reverse(self): self._data.reverse()

    def sort(self, *args, **kwds): self._data.sort(*args, **kwds)


class Dict(Hook, MutableMapping):
    def __init__(*args, **kwargs):
        if not args:
            raise TypeError("descriptor '__init__' of 'Dict' object "
                            "needs an argument")

        self, *args = args

        if len(args) > 1:
            raise TypeError('expected at most 1 arguments, got %d' % len(args))

        if args:
            dict_ = args[0]

        elif 'dict' in kwargs:
            dict_ = kwargs.pop('dict')
            import warnings
            warnings.warn("Passing 'dict' as keyword argument is deprecated",
                          PendingDeprecationWarning, stacklevel=2)
        else:
            dict_ = None

        self._data = {}

        if dict_ is not None:
            self.update(dict_)
        if len(kwargs):
            self.update(kwargs)

    # all set action will be here
    def __setitem__(self, key, item):
        if key in self.keys():
            del self[key]

        self._before_add(key=key, item=item)
        self._data[key] = item
        self._after_add(key=key, item=item)

    # all del action will be here
    def __delitem__(self, key):
        item = self[key]
        self._before_del(key=key, item=item)
        del self._data[key]
        self._after_del(key=key, item=item)

    def __iter__(self):
        return iter(self._data)

    def __getitem__(self, key):
        if key in self._data:
            return self._data[key]
        if hasattr(self.__class__, "__missing__"):
            return self.__class__.__missing__(self, key)
        raise KeyError(key)

    def __len__(self): return len(self._data)

    def copy(self):
        if self.__class__ is Dict:
            return Dict(self._data.copy())
        import copy
        data = self._data
        try:
            self._data = {}
            c = copy.copy(self)
        finally:
            self._data = data
        c.update(self)
        return c



"""Model definitions for data about an academic institution."""
from queue import Queue
from typing import Any, List, Optional

import gator._vendor.hooky as hooky
import mongoengine
from mongoengine import Document, EmbeddedDocument, fields


class _SubInstitutionsList(hooky.List, List['Institution']):
    """A list of sub-institutions for a wrapped institution.

    This list is used to ensure that the parent of each sub-institution is
    updated when the list is modified. This is done by intercepting calls to
    `append` and `remove`.
    """

    # Private Instance Attributes:
    #   _institution: The wrapped institution whose sub-institutions this list
    #       represents.
    _institution: 'Institution'

    def __init__(self, institution: 'Institution', **kwargs: Any) -> None:
        """Initialize a list of sub-institutions.

        Args:
            institution: The wrapped institution whose sub-institutions this
                list represents.
        """
        self._institution = institution
        super().__init__(self._institution._sub_institutions, **kwargs)

    def _after_add(self, key: Any, item: Any) -> None:
        """Call after an item is added to the list.

        This will set the parent of the added institution to the institution
        that this list was created for.
        """
        super()._after_add(key, item)
        assert isinstance(key, int), 'Only integer keys are allowed'
        assert isinstance(item, Institution), 'Only institutions are allowed'
        item.parent = self._institution

    def _after_del(self, key: Any, item: Any) -> None:
        """Call after an item is removed from the list.

        This will set the parent of the removed institution to None.
        """
        super()._after_del(key, item)
        assert isinstance(key, int), 'Only integer keys are allowed'
        assert isinstance(item, Institution), 'Only institutions are allowed'
        item.parent = None


class Institution(Document):
    """An academic institution.

    This is represented as a recursive tree structure, where each institution
    can have multiple sub-institutions. For example, the University of Toronto
    is the parent institution of the St. George campus of the University of
    Toronto, which is the parent institution of many faculties and departments.

    Instance Attributes:
        code: A unique string representing this institution.
        name: The full name of this institution.
        type: The type of this institution. For example, `university` or
            `campus`.
    """

    code: str = fields.StringField(unique=True, primary_key=True)  # type: ignore
    name: str = fields.StringField(required=True)  # type: ignore
    type: str = fields.StringField(required=True)  # type: ignore

    # Private Instance Attributes:
    #   _parent: The parent institution of this institution. For example, the
    #       University of Toronto is the parent institution of the St. George
    #       campus of the University of Toronto. If None, this institution is
    #       a top-level institution.
    #   _sub_institutions: A list of institutions that are sub-institutions of
    #       this institution. For example, the St. George campus of the
    #       University of Toronto is a sub-institution of the University of
    #       Toronto.
    _parent: Optional['Institution'] = fields.ReferenceField(
        'self',
        default=None,
        reverse_delete_rule=mongoengine.NULLIFY
    )  # type: ignore
    _sub_institutions: list['Institution'] = fields.ListField(
        fields.ReferenceField('self'),
        default=list,
        reverse_delete_rule=mongoengine.NULLIFY
    )  # type: ignore

    def __init__(self, **kwargs: Any) -> None:
        """Initialize an institution."""
        # Don't allow the user to set the parent or sub-institutions directly
        assert '_parent' not in kwargs, 'Use `parent` instead of `_parent`'
        assert '_sub_institutions' not in kwargs, \
            'Use `sub_institutions` instead of `_sub_institutions`'

        parent = kwargs.pop('parent', None)
        sub_institutions = kwargs.pop('sub_institutions', [])
        super().__init__(**kwargs)

        # Set the parent and sub-institutions
        self.parent = parent
        for sub_inst in sub_institutions:
            self.sub_institutions.append(sub_inst)

    @property
    def parent(self) -> Optional['Institution']:
        """Get the parent of this institution."""
        return self._parent

    @parent.setter
    def parent(self, new_parent: Optional['Institution']) -> None:
        """Set the parent of this institution.

        Args:
            new_parent: The new parent of this institution. If None, this
                institution will become a top-level institution.

        Examples:
            >>> foo = Institution(code='foo', name='Foo', type='university')
            >>> bar = Institution(code='bar', name='Bar', type='campus')
            >>> baz = Institution(code='baz', name='Baz', type='department')
            >>> bar.parent = foo
            >>> baz.parent = foo
            >>> bar.parent  # doctest: +ELLIPSIS
            Institution(code='foo', ...)
            >>> baz.parent  # doctest: +ELLIPSIS
            Institution(code='foo', ...)
            >>> foo.sub_institutions  # doctest: +ELLIPSIS
            [Institution(code='bar', ...), Institution(code='baz', ...)]
        """
        old_parent = self.parent
        if old_parent is not None:
            # Remove this institution from the old parent's list of
            # sub-institutions
            old_parent._sub_institutions.remove(self)

        self._parent = new_parent
        if new_parent is not None:
            # Add this institution to the new parent's list of sub-institutions
            new_parent._sub_institutions.append(self)

    @property
    def sub_institutions(self) -> _SubInstitutionsList:
        """Get the sub-institutions of this institution.

        >>> foo = Institution(code='foo', name='Foo', type='university')
        >>> bar = Institution(code='bar', name='Bar', type='campus')
        >>> baz = Institution(code='baz', name='Baz', type='department')
        >>> foo.sub_institutions.append(bar)
        >>> foo.sub_institutions.append(baz)
        >>> foo.sub_institutions  # doctest: +ELLIPSIS
        [Institution(code='bar', ...), Institution(code='baz', ...)]
        >>> bar.parent  # doctest: +ELLIPSIS
        Institution(code='foo', ...)
        >>> baz.parent  # doctest: +ELLIPSIS
        Institution(code='foo', ...)
        """
        return _SubInstitutionsList(self, hook_when_init=False)

    @sub_institutions.setter
    def sub_institutions(self, new_sub_institutions: list['Institution']) \
            -> None:
        """Set the sub-institutions of this institution.

        Args:
            new_sub_institutions: The new sub-institutions of this
                institution.
        """
        self._sub_institutions.clear()
        for sub_inst in new_sub_institutions:
            self.sub_institutions.append(sub_inst)

    def find_children(self, institution_type: str) -> list['Institution']:
        """Find all sub-institutions of the given type.

        This function recursively searches through all sub-institutions of this
        institution, performing a queue-based breadth-first search, and returns
        a list of all institutions of the given type.

        Args:
            institution_type: The type of institution to search for. For
                example, `university` or `campus`.

        Returns:
            A list of institutions of the given type.
        """
        q = Queue()
        q.put(self)

        visited = set()
        children = []

        while not q.empty():
            curr = q.get()
            if curr.code in visited:
                continue
            visited.add(curr.code)

            if curr.type == institution_type:
                children.append(curr)

            for sub_inst in curr.sub_institutions:
                q.put(sub_inst)

        return children

    def find_parent(self, institution_type: str) -> Optional['Institution']:
        """Find the first strictly parent institution of the given type.

        This function recursively searches through all parent institutions of
        this institution, performing a queue-based breadth-first search, and
        returns the first institution of the given type. This function does not
        include this institution in the search.

        Args:
            institution_type: The type of institution to search for. For
                example, `university` or `campus`.

        Returns:
            The first (closest) parent institution of the given type, or None
            if no such institution exists.
        """
        q = Queue()
        q.put(self)

        visited = set()

        while not q.empty():
            curr = q.get()
            if curr.code in visited:
                continue
            visited.add(curr.code)

            if curr.type == institution_type and curr != self:
                return curr

            if curr.parent is not None:
                q.put(curr.parent)

        return None

    def print_hierarchy(self) -> None:
        """Visualize this institution's hierarchy as a tree.

        This function prints a tree representation of this institution's
        hierarchy starting from this institution.

        Examples:
            >>> uoft = Institution(code='uoft', name='University of Toronto', type='university')
            >>> stgeorge = Institution(code='stgeorge', name='St. George', type='campus', parent=uoft)
            >>> uoft.print_hierarchy()
            University of Toronto (uoft, university)
            └── St. George (stgeorge, campus)

            >>> artsci = Institution(code='artsci', name='Faculty of Arts and Science', type='faculty', parent=stgeorge)
            >>> cs = Institution(code='cs', name='Department of Computer Science', type='department', parent=artsci)
            >>> uoft.print_hierarchy()
            University of Toronto (uoft, university)
            └── St. George (stgeorge, campus)
                └── Faculty of Arts and Science (artsci, faculty)
                    └── Department of Computer Science (cs, department)

            >>> utm = Institution(code='utm', name='Mississauga', type='campus', parent=uoft)
            >>> uoft.print_hierarchy()
            University of Toronto (uoft, university)
            ├── St. George (stgeorge, campus)
            │   └── Faculty of Arts and Science (artsci, faculty)
            │       └── Department of Computer Science (cs, department)
            └── Mississauga (utm, campus)

            >>> mat_at_utm = Institution(code='mat_at_utm', name='Department of Mathematics', type='department', parent=utm)
            >>> uoft.print_hierarchy()
            University of Toronto (uoft, university)
            ├── St. George (stgeorge, campus)
            │   └── Faculty of Arts and Science (artsci, faculty)
            │       └── Department of Computer Science (cs, department)
            └── Mississauga (utm, campus)
                └── Department of Mathematics (mat_at_utm, department)
        """
        def _helper(institution: 'Institution', prefix: str = '',
                    is_last: bool = True) -> None:
            """Print the hierarchy starting from the given institution.

            Args:
                institution: The root of the institution hierarchy.
                prefix: The prefix to print before the root institution's name.
                is_last: Whether this institution is the last child of its
                    parent.
            """
            has_parent = institution.parent is not None
            if has_parent:
                print(f'{prefix}└── ' if is_last else f'{prefix}├── ', end='')
                prefix += '    ' if is_last else '│   '

            print(str(institution))

            for i, sub_inst in enumerate(institution.sub_institutions):
                _helper(
                    sub_inst,
                    prefix=prefix,
                    is_last=i == len(institution.sub_institutions) - 1
                )

        _helper(self)

    def __str__(self) -> str:
        """Return a string representation of this institution.

        Returns:
            A string representation of this institution.
        """
        return f'{self.name} ({self.code}, {self.type})'

    def __repr__(self) -> str:
        """Return a string representation of this institution.

        Returns:
            A string representation of this institution.

        Examples:
            >>> uoft = Institution(code='uoft', name='University of Toronto', type='university')
            >>> repr(uoft)
            "Institution(code='uoft', name='University of Toronto', type='university', parent=None, sub_institutions=[])"
        """
        parent_code = self.parent.code if self.parent is not None else None
        sub_inst_codes = [inst.code for inst in self.sub_institutions]
        return f'Institution(code={self.code!r}, name={self.name!r}, ' \
            f'type={self.type!r}, parent={parent_code!r}, ' \
            f'sub_institutions={sub_inst_codes!r})'


class Building(Document):
    """A building on a campus of the University of Toronto.

    Instance Attributes:
        code: A unique string code that identifies this building.
        institution: The institution that this building belongs to.
        name: The name of this building. Can be None if unknown.
        map_url: The URL of this building's map. Can be None if unknown.
    """

    code: str = fields.StringField(unique=True, primary_key=True)  # type: ignore
    institution: Institution = fields.ReferenceField(Institution, required=True)  # type: ignore
    name: Optional[str] = fields.StringField(null=True, default=None)  # type: ignore
    map_url: Optional[str] = fields.URLField(null=True, default=None)  # type: ignore


class Location(EmbeddedDocument):
    """A location on campus represented as a building-room pair.

    Instance Attributes:
        building: The building that the section meets in.
        room: The room that the section meets in, not including the building
            code.
    """

    building: Building = fields.ReferenceField(Building, required=True)  # type: ignore
    room: str = fields.StringField(required=True)  # type: ignore
